/****************************************************************************/
/* do -- perform built in macro functions                                   */
/*--------------------------------------------------------------------------*/
/* copyright 1984, 1986, 1987  Michael M Rubenstein                         */
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#if defined _MSC_VER || defined __WATCOMC__
#include <process.h>
#endif

#if defined _MSC_VER || defined __WATCOMC__
#include <io.h>
#endif

#include "m4.h"

static struct defentry  *deftable = NULL;
static struct defentry  *defhash[HASHSIZE];

static void             mk_div_name(char *name, int num);
static void             undiv(int i);
static void             prtdef(char *name, char *def);
static int              hash(char *s);
static long             atonum(char *s);

#if defined __WATCOMC__
static char             *mktemp(char *p);
#endif

/****************************************************************************/
/* install a definition in the table                                        */
/****************************************************************************/
void dodef(char **afrom, char **ato)
{
  if (ato > afrom)
  {
    if (strcmp(afrom[0], afrom[1]) == 0)
    {
      fprintf(stderr, "%s: %s defined as self\n.", pgmname, afrom[0]);
      exit(EXIT_FAILURE);
    }
    install(afrom[0], afrom[1], StringType);
  }
  else
    install(afrom[0], nullstr, StringType);
}

/****************************************************************************/
/* select argument                                                          */
/****************************************************************************/
void doif(char **afrom, char **ato)
{
  for (;;)
  {
    if (ato - afrom < 2)
      return;
    if (strcmp(afrom[0], afrom[1]) == 0)
    {
      pbstr(afrom[2]);
      return;
    }
    if ((afrom += 3) > ato)
      return;
    if (afrom == ato)
    {
      pbstr(afrom[0]);
      return;
    }
  }
}

/****************************************************************************/
/* increment argument                                                       */
/****************************************************************************/
void doincr(char *num, int i)
{
  if (*num == EOS)
  {
    putbak('1');
    if (i < 0)
      putbak('-');
  }
  else
    pbnum(atonum(num) + i);
}

/****************************************************************************/
/* select substring                                                         */
/****************************************************************************/
void dosub(char **afrom, char **ato)
{
  int                   i, n, l;
  char                  *s, *t;

  s = afrom[0];
  if (afrom >= ato)
    return;
  l = strlen(s);
  i = (int) atonum(afrom[1]);
  if (i > l)
    return;
  s += i;
  if (afrom + 1 == ato)
  {
    pbstr(s);
    return;
  }
  n = (int) atonum(afrom[2]);
  if (n <= 0)
    return;
  if (i + n > l)
  {
    pbstr(s);
    return;
  }
  for (t = s + n; --t >= s;)
    putbak(*t & 0xff);
}

/****************************************************************************/
/* change quote character                                                   */
/****************************************************************************/
void dochq(char **afrom, char **ato)
{
  if (afrom == ato && *afrom[0] == EOS)
  {
    lquote = LQUOTE;
    rquote = RQUOTE;
    return;
  }
  lquote = *afrom[0] & 0xff;
  if (afrom < ato)
    ++afrom;
  rquote = *afrom[0] & 0xff;
}

/****************************************************************************/
/* change parentheses characters                                            */
/****************************************************************************/
void doparen(char **afrom)
{
  if (*afrom[0] == EOS)
  {
    lparen = LPAREN;
    rparen = RPAREN;
  }
  else
  if (strlen(afrom[0]) > 1)
  {
    lparen = afrom[0][0] & 0xff;
    rparen = afrom[0][1] & 0xff;
  }
}

/****************************************************************************/
/* change comma character                                                   */
/****************************************************************************/
void docomma(char **afrom)
{
  if (*afrom[0] == EOS)
    comma = COMMA;
  else
    comma = *afrom[0] & 0xff;
}

/****************************************************************************/
/* set alternate quote characters                                           */
/****************************************************************************/
void doaquot(char **afrom, char **ato)
{
  int                   n;
  char                  *p;

  p = aquotes;
  for (n = AQUOTSZ; n-- && afrom <= ato && *afrom[0] != EOS;)
  {
    *(p++) = **afrom;
    if (afrom < ato)
      ++afrom;
    *p = (char) ((**afrom == EOS) ? *(p - 1) : **afrom);
    ++p;
    ++afrom;
  }
  *p = '\0';
}

/****************************************************************************/
/* skip to end of line                                                      */
/****************************************************************************/
void dodnl(void)
{
  int                   c;

  while ((c = ngetc()) != '\n' && c != EOF)
    ;
}

void domktmp(char *s)
{
  char                  *p;

  if ((p = mktemp(s)) != NULL)
    pbstr(p);
}

void dogetenv(char *s)
{
  char              *p;

  if ((p = getenv(s)) != NULL)
    pbstr(p);
}

void doignorenl(char *s)
{
  if (*s == EOS)
    putbak(ignorenl + '0');
  else
    ignorenl = (atonum(s) != 0l);
}

void dosyscmd(char *s)
{
#if !defined SUN
/*--------------------------------------------------------------------------*/
/* Sun doesn't handle fflush(NULL) correctly (may be fixed in new versions  */
/*--------------------------------------------------------------------------*/
  fflush(NULL);
#endif

  if (*s == EOS)
  {
#if defined MSDOS_SPAWN
    char *cmd = getenv("COMSPEC");

    if (cmd == NULL)
      cmd = "c:\\command.com";

    spawnl(P_WAIT, cmd, cmd, (char *) NULL);
#elif defined UNIX
    char *cmd = getenv("SHELL");

    if (cmd == NULL)
      cmd = "/bin/sh";

    system(cmd);
#endif
  }
  else
    system(s);
}

/****************************************************************************/
/* output a character specified as number                                   */
/****************************************************************************/
void dochar(char **afrom, char **ato)
{
  while (afrom <= ato)
  {
    if (*afrom[0] != EOS)
      putbak((int) atonum(*afrom) & 0xff);
    ++afrom;
  }
}

/****************************************************************************/
/* test if symbol defined                                                   */
/****************************************************************************/
void doifdef(char **afrom, char **ato)
{
  if (afrom >= ato)
    return;
  if (lookup(afrom[0]))
    pbstr(afrom[1]);
  else
    if (afrom + 2 <= ato)
      pbstr(afrom[2]);
}

void doindex(char **afrom, char **ato)
{
  char                  *p;

  if (afrom >= ato)
  {
    putbak('0');
    return;
  }
  pbnum(((p = strstr(afrom[0], ato[0])) == NULL) ? -1L
                                                 : (long) (p - afrom[0]));
}

#if defined _MSC_VER
#pragma warning(disable : 4701)
#endif

/****************************************************************************/
/* transliterate                                                            */
/****************************************************************************/
void dotrans(char **afrom, char **ato)
{
  char                  *s, *s0, *from, *to;
  int                   lto;
  int                   c;
  int                   i;

  if (afrom >= ato)
    return;
  s0 = afrom[0];
  from = afrom[1];
  if (afrom + 2 > ato)
    lto = 0;
  else
    lto = strlen(to = afrom[2]);
  for (s = s0 + strlen(s0) - 1; s >= s0; --s)
  {
    c = *s;
    for (i = 0; from[i] && c != from[i]; ++i)
      ;
    if (from[i])
    {
      if (i < lto)
        putbak(to[i] & 0xff);
    }
    else
      putbak(c & 0xff);
  }
}

#if defined _MSC_VER
#pragma warning(default: 4701)
#endif

/****************************************************************************/
/* delete a definition                                                      */
/****************************************************************************/
void doundef(char *name)
{
  struct defentry       *p, *q, *d;
  int                   n;

  for (d = defhash[n = hash(name)], p = NULL;
       d != NULL && strcmp(d->name, name) != 0;
       p = d, d = d->hnext)
    ;

  if (d != NULL)
  {
    if (p == NULL)
      defhash[n] = d->hnext;
    else
      p->hnext = d->hnext;

    for (p = NULL, q = deftable; q != d; p = q, q = q->next)
      ;

    if (p == NULL)
      deftable = d->next;
    else
      p->next = d->next;

    free(d->name);
    free(d);
  }
}

/****************************************************************************/
/* delete builtin definitions                                               */
/****************************************************************************/
void donobuilt(void)
{
  struct defentry       *p, *q, *r;

  for (q = NULL, p = deftable; p != NULL;)
    if (p->type)
    {
      r = p;
      if (q == NULL)
        p = deftable = p->next;
      else
        p = q->next = p->next;
      free(r->name);
      free(r);
    }
    else
    {
      q = p;
      p = p->next;
    }
  doundef("msdos");

 dodnl();
}

/****************************************************************************/
/* include a file                                                           */
/****************************************************************************/
void doincl(char *name, int quiet)
{
  if (*name == EOS)
    return;
  if (++isp >= istk + ISTSIZE)
    error("include nesting too deep");
  if (!newin(name, quiet))
    --isp;
} 

/****************************************************************************/
/* divert output                                                            */
/****************************************************************************/
void dodiv(char **divno)
{
  char                  outname[100];

  curout = (int) atonum(*divno);
  if (curout< 0 || curout > 9)
  {
    curout = -1;
    return;
  }
  if (outfile[curout] == NULL)
  {
    mk_div_name(outname, curout);
    if ((outfile[curout] = fopen(outname, "w")) == NULL)
      error("cannot create diversion file");
  }
}

/****************************************************************************/
/* put diversion to output                                                  */
/****************************************************************************/
void doundiv(char **afrom, char **ato)
{
  if (afrom == ato && *afrom[0] == EOS)
    undivall();
  else
    while (afrom <= ato)
      undiv((int) atonum(*(afrom++)));
}

/****************************************************************************/
/* undivert all diversion files                                             */
/****************************************************************************/
void undivall(void)
{
  int                   i;

  for (i = 1; i <= 9; ++i)
    undiv(i);
}

/****************************************************************************/
/* undivert one diversion file                                              */
/****************************************************************************/
static void undiv(int i)
{
  FILE                  *f;
  int                   c;
  char                  outname[64];

  if (i == curout || i < 1 || i > 9 || outfile[i] == NULL)
    return;
  fclose(outfile[i]);
  mk_div_name(outname, i);
  if (curout >= 0)
  {
    if ((f = fopen(outname, "r")) == NULL)
      error("cannot open diversion file");
    while ((c = getc(f)) != EOF)
      putc(c, outfile[curout]);
    fclose(f);
  }
  unlink(outname);
  outfile[i] = NULL;
}

static void mk_div_name(char *name, int num)
{
  char                  *p;

  if ((p = getenv("TMP")) != NULL && *p != '\0')
  {
    strcpy(name, p);
    name += strlen(name);
    if (*(name - 1) != '\\' && *(name - 1) != ':')
      *(name++) = '\\';
  }
  sprintf(name, "m4temp%d.dat", num);
}

/****************************************************************************/
/* print out definitions                                                    */
/****************************************************************************/
void dodump(char **afrom, char **ato)
{
  struct defentry       *d;

  if (afrom == ato && !*afrom[0])
    for (d = deftable; d != NULL; d = d->next)
      prtdef(d->name, d->def);
  else
    while (afrom <= ato)
    {
      if ((d = lookup(afrom[0])) != NULL)
        prtdef(afrom[0], d->def);
      ++afrom;
    }
}

/****************************************************************************/
/* print out a definition                                                   */
/****************************************************************************/
static void prtdef(char *name, char *def)
{
  fprintf(stderr, "%c%s%c\t%c%s%c\n", lquote, name, rquote, lquote,
                  def, rquote);
}

/****************************************************************************/
/* change or report file name                                               */
/****************************************************************************/
void setoutput(char *fname, char *append, char *curname, FILE **f)
{
  if (*fname == '\0')
    pbstr(curname);
  else
  {
    strncpy(curname, fname, FILENAME_MAX);
    curname[FILENAME_MAX - 1] = '\0';
    if (*f != stdout)
      fclose(*f);
    if (strcmp(curname, "-") == 0)
      *f = stdout;
    else
    if (strcmp(curname, "--") == 0)
      *f = stderr;
    else
      if ((*f = fopen(curname, (*append == 'a') ? "a" : "w"))
              == NULL)
        error("cannot create file %s", curname);
  }
}

/****************************************************************************/
/* set trace on                                                             */
/****************************************************************************/
void dotrace(char **afrom, char **ato)
{
  struct defentry       *d;

  while (afrom <= ato)
  {
    if ((d = lookup(afrom[0])) != NULL)
      d->trace = TRUE;
    ++afrom;
  }
}

/****************************************************************************/
/* set trace off                                                            */
/****************************************************************************/
void dountrace(char **afrom, char **ato)
{
  struct defentry       *d;

  if (afrom == ato && !*afrom[0])
    for (d = deftable; d != NULL; d = d->next)
      d->trace = FALSE;
  else
    while (afrom <= ato)
    {
      if ((d = lookup(afrom[0])) != NULL)
        d->trace = FALSE;
      ++afrom;
    }
}

/****************************************************************************/
/* install a definition                                                     */
/****************************************************************************/
void install(char *name, char *defn, unsigned type)
{
  int                   n1, n2;
  struct defentry       *d;

  if (!isalpha(*name) && *name != '_')
    return;
  n1 = strlen(name) + 1;
  n2 = strlen(defn) + 1;

  if ((d = malloc(sizeof(struct defentry))) == NULL
   || (d->name = malloc(n1 + n2)) == NULL)
    error("insufficient memory.");

  d->type = type;
  d->trace = FALSE;
  strcpy(d->name, name);
  strcpy(d->def = d->name + n1, defn);
  d->next = deftable;
  deftable = d;

  d->hnext = defhash[n1 = hash(name)];
  defhash[n1] = d;
}

/****************************************************************************/
/* lookup a definition                                                      */
/****************************************************************************/
struct defentry *lookup(char *name)
{
  struct defentry       *d;

  for (d = defhash[hash(name)]; d != NULL; d = d->hnext)
    if (strcmp(name, d->name) == 0)
      return d;

  return NULL;
}

/****************************************************************************/
/* compute hash index of name                                               */
/****************************************************************************/
static int hash(char *s)
{
#if HASHSIZE < 256
  unsigned              h;
#else
  long                  h;
#endif

  h = 0;
  while (*s != '\0')
    h = ((h << 8) + (*(s++) & 0xff)) %  HASHSIZE;

  return (int) h;
}

/****************************************************************************/
/* convert string to long with error checking                               */
/****************************************************************************/
static long atonum(char *s)
{
  long                  n;
  int                   minus;

  n = minus = 0;
  while (*s == '+' || *s == '-')
  {
    if (*s == '-')
      minus = !minus;
    ++s;
  }
  while (isdigit(*s))
    n = n * 10 + (*(s++) - '0');
  while (isspace(*s))
    ++s;
  if (*s != '\0')
    error("invalid number");
  return minus ? -n : n;
}

#if defined __WATCOMC__
char *mktemp(char* p)
{
  unsigned pid = getpid();
  char* q = p + strlen(p);
  int uniq;

  while (q > p && *(q - 1) == 'X')
  {
    *--q = (char) ('0' + pid % 10);
    pid /= 10;
  }

  if (*q == '\0')
    return NULL;

  for (uniq = 'a'; uniq <= 'z'; ++uniq)
  {
    *q = (char) uniq;
    if (access(p, F_OK))
      return p;
  }
  return NULL;
}
#endif
