/****************************************************************************/
/* m4 -- macro processor                                                    */
/* copyright 1984, 88, 91, 92  Michael M Rubenstein                         */
/****************************************************************************/

/****************************************************************************/
/* 14 Jul 94    1.71    Added support for Sun                               */
/*                                                                          */
/* 22 Jun 94    1.7     Removed swapcmd.  Added support for Watcom C++.     */
/*                                                                          */
/* 14 Apr 93    1.61    Fixed bug in -o option                              */
/*                                                                          */
/* 13 Jan 92    1.6     Added hashing of defined names for faster lookup    */
/*                                                                          */
/* 13 Dec 91    1.5     Added output, redefine, trace, changecomma,         */
/*                      changeparen.  Removed emit.                         */
/*                                                                          */
/* 24 Jan 91    1.4     Added nulout, char, swapcmd.                        */
/*                      Fixed minor bugs.                                   */
/*                                                                          */
/* 21 Aug 88    1.3     Modified to allow input with high bit set.          */
/*                      Fixed bug in undefine and nobuiltin.                */
/*                                                                          */
/* 22 May 88    1.2     Added printf macro.                                 */
/*                                                                          */
/* 16 Jan 87    1.1     Added emit and getenv macros.  Modifed diversion    */
/*                      files to use directory specified by TMP environment */
/*                      variable.                                           */
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>

#include "m4.h"

#if defined DOS
#include "util.h"
#endif

#define PGMNAME         "m4"

char                    *evals;
char                    *ep;

struct callstack        *cp = NULL;

char                    *argstack[ARGSIZE];
char                    **ap = argstack;
char                    *ibuf;
char                    *ip;
struct istack           istk[ISTSIZE];
struct istack           *isp = istk;

FILE                    *outfile[10] = { NULL, NULL, NULL, NULL, NULL,
                                         NULL, NULL, NULL, NULL, NULL };
FILE                    *tracefile = stderr;
int                     curout = 0;
char                    aquotes[2 * AQUOTSZ + 1] = "";
char                    nullstr[] = "";
char                    *pgmname = PGMNAME;

int                     ignorenl = 0;
int                     lquote = LQUOTE,
                        rquote = RQUOTE,
                        lparen = LPAREN,
                        rparen = RPAREN,
                        comma = COMMA,
                        argflag = ARGFLAG,
                        comment = COMMENT,
                        macchr = 0,
                        nulout = 0;

static int              wargc;
static char             **wargv;

static char             outfilename[FILENAME_MAX] = "-";
static char             tracefilename[FILENAME_MAX] = "--";

static void             eval(MacroType type, unsigned trace,
                             char **afrom, char **ato);
static void             skipsp(void);
static void             help(void);

int main(int argc, char **argv)
{
  static struct callstack
                        callst[CALLSIZE];
  struct defentry       *defn;
  static char           token[MAXTOKEN + 1];
  char                  *tk;
  int                   t;
  int                   nlb;

  if ((ep = evals = malloc(EVALSIZE)) == NULL
   || (ip = ibuf = malloc(IBUFSIZ)) == NULL)
    error("insuf mem");

  outfile[0] = stdout;

  wargc = argc;
  wargv = argv;
  while ((t = getopt(argc, argv, "o:?")) != EOF)
    switch (t)
    {
      case 'o':         if (optarg[0] == '\0')
                        {
                          fprintf(stderr, "%s: output file name missing\n",
                                  pgmname);
                          exit(EXIT_FAILURE);
                        }
                        setoutput(optarg, "", outfilename, &outfile[0]);
                        break;

      default:          help();
                        exit(EXIT_FAILURE);
    }

  install("untrace", nullstr, UntraceType);
#if defined UNIX
  install("unix", nullstr, StringType);
#endif  
  install("undivert", nullstr, UndivType);
  install("undefine", nullstr, UndType);
  install("translit", nullstr, TransType);
  install("tracefile", nullstr, TracefileType);
  install("trace", nullstr, TraceType);
  install("syscmd", nullstr, SyscmdType);
  install("substr", nullstr, SubType);
  install("sinclude", nullstr, SinclType);
  install("redefine", nullstr, RedefType);
  install("printf", nullstr, PrintfType);
  install("output", nullstr, OutputType);
  install("nulout", nullstr, NuloutType);
  install("nobuiltin", nullstr, NobuiltType);
#if defined MSDOS
  install("msdos", nullstr, StringType);
#endif  
  install("maketemp", nullstr, MktmpType);
  install("macro", nullstr, MacType);
  install("len", nullstr, LenType);
  install("index", nullstr, IndType);
  install("incr", nullstr, IncrType);
  install("include", nullstr, InclType);
  install("ignorenl", nullstr, IgnorenlType);
  install("ifelse", nullstr, IfType);
  install("ifdef", nullstr, IfdType);
  install("getenv", nullstr, GetenvType);
  install("eval", nullstr, EvType);
  install("errprint", nullstr, ErrType);
  install("dumpdef", nullstr, DumpType);
  install("dnl", nullstr, DnlType);
  install("divnum", nullstr, DivnType);
  install("divert", nullstr, DivType);
  install("define", nullstr, DefType);
  install("decr", nullstr, DecrType);
  install("comment", nullstr, CmntType);
  install("changequote", nullstr, ChqType);
  install("changeparen", nullstr, ParenType);
  install("changecomma", nullstr, CommaType);
  install("char", nullstr, CharType);
  install("changearg", nullstr, ChargType);
  install("aquote", nullstr, AquoType);

  if (optind >= wargc)
  {
    istk[0].ifile = stdin;
    istk[0].ip = ibuf;
  }
  else
    nextfile();

  while ((t = gettoken(token, MAXTOKEN)) != EOF)
  {
    switch (t)
    {
      case WORD:        if (macchr == 0)
                        {
                          if ((defn = lookup(tk = token)) == NULL)
                          {
                            puttoken(token);
                            break;
                          }
                        }
                        else
                        {
                          if ((defn = lookup(tk = token + 1)) == NULL)
                          {
                            puttoken(token);
                            break;
                          }
                        }

                        cp = (cp == NULL) ? callst : ++cp;
                        if (cp >= callst + CALLSIZE)
                          error("call stack overflow.");
                        cp->args = ap;
                        cp->type = defn->type;
                        cp->trace = defn->trace;
                        push();
                        puttoken(defn->def);
                        putchr(EOS);
                        push();
                        puttoken(tk);
                        putchr(EOS);
                        push();
                        putbak(t = ngetc());
                        if (t != lparen)
                        {
                          putbak(rparen);
                          putbak(lparen);
                        }
                        cp->plev = 0;
                        break;

      case LQUOTE:      nlb = 1;
                        do
                        {
                          if ((t = ngetc()) == EOF)
                            error("EOF in string.");
                          if (t == rquote)
                            --nlb;
                          else
                          if (t == lquote)
                            ++nlb;
                          if (nlb)
                            putchr(t);
                        } while (nlb);
                        break;

      case COMMENT:     puttoken(token);
                        while ((t = ngetc()) != '\n' && t != EOF)
                          putchr(t);
                        if (t == '\n')
                          putchr('\n');
                        break;

      default:          if (cp == NULL)
                        {
                          puttoken(token);
                          break;
                        }
                        switch (t)
                        {
                          case LPAREN:    if ((cp->plev)++)
                                          {
                                            puttoken(token);
                                            break;
                                          }
                                          skipsp();
                                          break;

                          case RPAREN:    if (--(cp->plev))
                                          {
                                            puttoken(token);
                                            break;
                                          }
                                          putchr(EOS);
                                          eval(cp->type, cp->trace,
                                               cp->args, ap - 1);
                                          ep = *(ap = cp->args);
                                          cp = (cp > callst) ? --cp
                                                             : NULL;
                                          break;

                          case COMMA:     if (cp->plev == 1)
                                          {
                                            putchr(EOS);
                                            push();
                                            skipsp();
                                            break;
                                          }
                                          puttoken(token);
                                          break;

                          default:        puttoken(token);
                          }
    }
  }

  if (cp != NULL)
    error("unexpected EOF");
  curout = 0;
  undivall();

  if (outfile[0] != stdout && outfile[0] != stderr)
    fclose(outfile[0]);
  if (tracefile != stdout && tracefile != stderr)
    fclose(tracefile);

  return EXIT_SUCCESS;
}

/****************************************************************************/
/* put a token to output or evaluation stack as appropriate                 */
/****************************************************************************/
void puttoken(char *s)
{
  while (*s)
    putchr(*(s++));
}

/****************************************************************************/
/* put a character to output or evaluation stack as appropriate             */
/****************************************************************************/
void putchr(int c)
{
  if (cp == NULL)
  {
    if (c != 0 && c != nulout)
      if (curout >= 0)
        putc(c, outfile[curout]);
  }
  else
  {
    if (ep >= evals + EVALSIZE)
      error("evaluation stack overflow.");
    *(ep)++ = (char) c;
  }
}

/****************************************************************************/
/* push ep onto argstack, return updated pointer                            */
/****************************************************************************/
void push(void)
{
  if (ap >= argstack + ARGSIZE)
    error("arg stack overflow.");
  *(ap++) = ep;
}

/****************************************************************************/
/* expand args afrom through ato; evaluate builtin or push back definition  */
/****************************************************************************/
static void eval(MacroType type, unsigned trace, char **afrom, char **ato)
{
  char                  *def, *s;
  char                  **arg;

  if (trace)
  {
    fprintf(tracefile, "\n***tracing %s\n", afrom[1]);
    for (arg = afrom + 2; arg <= ato; ++arg)
      fprintf(tracefile, "           %s\n", *arg);
    fputs("***\n\n", tracefile);
  }  

  afrom += 2;
  switch (type)
  {
    case DefType:     dodef(afrom, ato);
                      break;

    case IfType:      doif(afrom, ato);
                      break;

    case IncrType:    doincr(afrom[0], 1);
                      break;

    case DecrType:    doincr(afrom[0], -1);
                      break;

    case SubType:     dosub(afrom, ato);
                      break;

    case EvType:      doeval(afrom[0]);
                      break;

    case ChqType:     dochq(afrom, ato);
                      break;

    case UndType:     doundef(afrom[0]);
                      break;

    case IfdType:     doifdef(afrom, ato);
                      break;

    case InclType:    doincl(afrom[0], FALSE);
                      break;

    case SinclType:   doincl(afrom[0], TRUE);
                      break;

    case DivType:     dodiv(afrom);
                      break;

    case UndivType:   doundiv(afrom, ato);
                      break;

    case DivnType:    pbnum((long) curout);
                      break;

    case LenType:     pbnum((long) strlen(afrom[0]));
                      break;

    case IndType:     doindex(afrom, ato);
                      break;

    case TransType:   dotrans(afrom, ato);
                      break;

    case ErrType:     fprintf(stderr, afrom[0], afrom[1], afrom[2],
                              afrom[3], afrom[4], afrom[5], afrom[6],
                              afrom[7], afrom[8]);
                      break;

    case PrintfType:  if (curout >= 0)
                        fprintf(outfile[curout], afrom[0], afrom[1],
                                afrom[2], afrom[3], afrom[4], afrom[5],
                                afrom[6], afrom[7], afrom[8]);
                      break;

    case DumpType:    dodump(afrom, ato);
                      break;

    case DnlType:     dodnl();
                      break;

    case ChargType:   argflag = (*afrom[0] == EOS) ? ARGFLAG
                                                   : *afrom[0] & 0xff;
                      break;

    case AquoType:    doaquot(afrom, ato);
                      break;

    case NobuiltType: donobuilt();
                      break;

    case CmntType:    comment = *afrom[0] & 0xff;
                      break;

    case MacType:     macchr = *afrom[0] & 0xff;
                      break;

    case NuloutType:  nulout = *afrom[0] & 0xff;
                      break;

    case MktmpType:   domktmp(afrom[0]);
                      break;

    case SyscmdType:  dosyscmd(afrom[0]);
                      break;

    case CharType:    dochar(afrom, ato);
                      break;

    case GetenvType:  dogetenv(afrom[0]);
                      break;

    case IgnorenlType:
                      doignorenl(afrom[0]);
                      break;

    case OutputType:  setoutput(afrom[0], (ato > afrom) ? afrom[1] : "",
                                outfilename, &outfile[0]);
                      break;

    case RedefType:   doundef(afrom[0]);
                      dodef(afrom, ato);
                      break;

    case TraceType:   dotrace(afrom, ato);
                      break;

    case UntraceType: dountrace(afrom, ato);
                      break;

    case TracefileType:
                      setoutput(afrom[0], (ato > afrom) ? afrom[1] : "",
                                tracefilename, &tracefile);
                      break;

    case ParenType:   doparen(afrom);
                      break;

    case CommaType:   docomma(afrom);
                      break;

    case StringType:  afrom -= 2;
                      def = afrom[0];
                      for (s = def + strlen(def); --s > def;)
                      {
                        if ((*(s - 1) & 0xff) != argflag)
                          putbak(*s & 0xff);
                        else
                        {
                          if (isascii(*s) && isdigit(*s))
                          {
                            arg = afrom + *s - ('0' - 1);
                            if (arg <= ato)
                              pbstr(*arg);
                          }
                          --s;
                        }
                      }
                      if (s == def)
                        putbak(*s & 0xff);
                      break;
  }
}

/****************************************************************************/
/* convert number to string and push back                                   */
/****************************************************************************/
void pbnum(long n)
{
  int                   minus;

  minus = FALSE;
  if (n < 0)
  {
    n = -n;
    minus = TRUE;
  }

  do
  {
    putbak((int) (n % 10l) + '0');
    n /= 10;
  } while (n != 0);
  if (minus)
    putbak('-');
}

/****************************************************************************/
/* skip white space                                                         */
/****************************************************************************/
static void skipsp(void)
{
  int                   c;

  while (isspace(c = ngetc()))
    ;
  putbak(c);
}

/****************************************************************************/
/* push characters back onto input                                          */
/****************************************************************************/
void putbak(int c)
{
  if (c == EOF || c == '\0')            /* don't push back EOF or NUL */
    return;
  if (ip >= ibuf + IBUFSIZ)
    error("too many characters pushed back.");
  *(ip++) = (char) c;
}

/****************************************************************************/
/* get a (possibly pushed back) character                                   */
/****************************************************************************/
int ngetc(void)
{
  int                   c;
  static int            endfile = FALSE;

  for (;;)
  {
    if (ip > isp->ip)
      return (*--ip) & 0xff;            /* pushed back character */
    if (endfile)
      return EOF;
    if ((c = getc(isp->ifile)) != EOF)
    {
      if (ignorenl && c == '\n')
        continue;
      if (c == '\0')
        continue;
      return c;
    }
    if (isp->ifile != stdin)
      fclose(isp->ifile);
    if (isp > istk)
      --isp;
    else
      if (!nextfile())
      {
        endfile = TRUE;
        return 0;                       /* make sure token ended before EOF */
      }
  }
}

/****************************************************************************/
/* switch to the next file                                                  */
/****************************************************************************/
int nextfile(void)
{
  if (optind >= wargc)
    return FALSE;
  newin(wargv[optind++], FALSE);
  return TRUE;
}

/****************************************************************************/
/* open an input file                                                       */
/****************************************************************************/
int newin(char *name, int quiet)
{
  isp->ip = ip;                       /* save pos on push back stack */
  if (strcmp(name, "-") == 0)
    isp->ifile = stdin;
  else
  if ((isp->ifile = fopen(name, "r")) == NULL)
  {
    if (quiet)
      return FALSE;
    fprintf(stderr, "%s: cannot open %s.\n", pgmname, name);
    exit(EXIT_FAILURE);
  }
  return TRUE;
}

/****************************************************************************/
/* push a string back                                                       */
/****************************************************************************/
void pbstr(char *s)
{
  char                  *t;

  for (t = s + strlen(s) - 1; t >= s; --t)
    putbak(*t & 0xff);
}

/****************************************************************************/
/* display error message                                                    */
/****************************************************************************/
void error(char *f, ...)
{
  va_list               ap;

  fprintf(stderr, "%s: ", pgmname);
  va_start(ap, f);
  vfprintf(stderr, f, ap);
  va_end(ap);

  putc('\n', stderr);

  exit(EXIT_FAILURE);
}

/****************************************************************************/
/* display help message                                                     */
/****************************************************************************/
void help(void)
{
  fprintf(stderr, "syntax:  %s [-o<outfile>] [<infile> ...]\n", pgmname);
}
